From 2f5d4386ba7e20db8d4dcce2ba3b73aaca5eca52 Mon Sep 17 00:00:00 2001
From: "Chao Li (Evan)" <lic@highgo.com>
Date: Tue, 27 Jan 2026 12:41:05 +0800
Subject: [PATCH v9] tablecmds: fix bug where index rebuild loses replica
 identity on partitions

ALTER TABLE ... ALTER COLUMN TYPE may require rebuilding dependent
indexes.  When the rebuilt index is marked as REPLICA IDENTITY on a
partitioned table, tablecmds previously failed to restore replica
identity on the affected partitions, leaving logical replication
misconfigured.

Fix this by tracking replica identity indexes using OIDs and by
recursively collecting replica identity indexes on all partitions of a
partitioned table.  After index rebuilds complete, restore replica
identity markings for each affected table.

Add regression tests covering multi-level partition hierarchies,
including partitions in different schemas, to verify that replica
identity is preserved across index rebuilds.

Author: Chao Li <lic@highgo.com>
Reviewed-by: Michael Paquier <michael@paquier.xyz>
Reviewed-by: Xuneng Zhou <xunengzhou@gmail.com>
Discussion: https://postgr.es/m/DB533C25-C6BA-4C0F-8046-96168E9CDD72@gmail.com
---
 src/backend/commands/tablecmds.c              | 135 +++++++++++++-----
 .../regress/expected/replica_identity.out     | 103 +++++++++++++
 src/test/regress/sql/replica_identity.sql     |  72 ++++++++++
 3 files changed, 272 insertions(+), 38 deletions(-)

diff --git a/src/test/regress/expected/replica_identity.out b/src/test/regress/expected/replica_identity.out
index 336b04fa278..c51f539dd45 100644
--- a/src/test/regress/expected/replica_identity.out
+++ b/src/test/regress/expected/replica_identity.out
@@ -292,6 +292,109 @@ ALTER TABLE test_replica_identity5 DROP CONSTRAINT test_replica_identity5_pkey;
 ERROR:  constraint "test_replica_identity5_pkey" of relation "test_replica_identity5" does not exist
 ALTER TABLE test_replica_identity5 ALTER b DROP NOT NULL;
 ERROR:  column "b" is in index used as replica identity
+--
+-- Test index rebuild preserves replica identity against partitioned tables
+--
+CREATE TABLE test_replica_identity_partitioned (id int NOT NULL, val int NOT NULL) PARTITION BY RANGE (id);
+CREATE TABLE test_replica_identity_partitioned_p1 PARTITION OF test_replica_identity_partitioned
+    FOR VALUES FROM (0) TO (100);
+CREATE TABLE test_replica_identity_partitioned_p2 PARTITION OF test_replica_identity_partitioned
+    FOR VALUES FROM (101) TO (200) PARTITION BY LIST (id);
+CREATE TABLE test_replica_identity_partitioned_p2_1 PARTITION OF test_replica_identity_partitioned_p2
+    FOR VALUES IN (150, 160);
+-- For better coverage, create a partition in a different schema
+CREATE SCHEMA test_replica_identity_schema;
+CREATE TABLE test_replica_identity_schema.test_replica_identity_partitioned_p2_2 PARTITION OF test_replica_identity_partitioned_p2
+    FOR VALUES IN (170, 180);
+CREATE UNIQUE INDEX test_replica_identity_partitioned_pkey ON test_replica_identity_partitioned (id, val);
+ALTER TABLE test_replica_identity_partitioned REPLICA IDENTITY USING INDEX test_replica_identity_partitioned_pkey;
+ALTER TABLE test_replica_identity_partitioned_p1 REPLICA IDENTITY USING INDEX test_replica_identity_partitioned_p1_id_val_idx;
+ALTER TABLE test_replica_identity_partitioned_p2_1 REPLICA IDENTITY USING INDEX test_replica_identity_partitioned_p2_1_id_val_idx;
+ALTER TABLE test_replica_identity_schema.test_replica_identity_partitioned_p2_2 REPLICA IDENTITY USING INDEX test_replica_identity_partitioned_p2_2_id_val_idx;
+SELECT tc.relname as table_name, tc.relreplident, i.indexrelid::regclass AS index_name, i.indisreplident FROM pg_class c
+  LEFT JOIN pg_index i ON c.oid = i.indexrelid
+  LEFT JOIN pg_class tc ON i.indrelid = tc.oid
+  WHERE c.relname LIKE 'test_replica_identity_partitioned%' and c.relkind in ('i', 'I')
+  ORDER BY c.relname;
+               table_name               | relreplident |                                   index_name                                   | indisreplident 
+----------------------------------------+--------------+--------------------------------------------------------------------------------+----------------
+ test_replica_identity_partitioned_p1   | i            | test_replica_identity_partitioned_p1_id_val_idx                                | t
+ test_replica_identity_partitioned_p2_1 | i            | test_replica_identity_partitioned_p2_1_id_val_idx                              | t
+ test_replica_identity_partitioned_p2_2 | i            | test_replica_identity_schema.test_replica_identity_partitioned_p2_2_id_val_idx | t
+ test_replica_identity_partitioned_p2   | d            | test_replica_identity_partitioned_p2_id_val_idx                                | f
+ test_replica_identity_partitioned      | i            | test_replica_identity_partitioned_pkey                                         | t
+(5 rows)
+
+-- Store the index state before rebuild
+SELECT array_agg(c.oid ORDER BY c.relname) AS index_oids,
+       array_agg(c.relname::text ORDER BY c.relname) AS index_names,
+       array_agg(n.nspname ORDER BY c.relname) AS schema_names,
+       array_agg(i.indisreplident ORDER BY c.relname) AS indisreplidents
+  FROM pg_class c
+  JOIN pg_namespace n ON c.relnamespace = n.oid
+  JOIN pg_index i ON c.oid = i.indexrelid
+  WHERE c.relname LIKE 'test_replica_identity_partitioned%' AND c.relkind IN ('i', 'I') \gset before_
+ALTER TABLE test_replica_identity_partitioned ALTER COLUMN val TYPE bigint;
+SELECT tc.relname as table_name, tc.relreplident, i.indexrelid::regclass AS index_name, i.indisreplident FROM pg_class c
+  LEFT JOIN pg_index i ON c.oid = i.indexrelid
+  LEFT JOIN pg_class tc ON i.indrelid = tc.oid
+  WHERE c.relname LIKE 'test_replica_identity_partitioned%' and c.relkind in ('i', 'I')
+  ORDER BY c.relname;
+               table_name               | relreplident |                                   index_name                                   | indisreplident 
+----------------------------------------+--------------+--------------------------------------------------------------------------------+----------------
+ test_replica_identity_partitioned_p1   | i            | test_replica_identity_partitioned_p1_id_val_idx                                | t
+ test_replica_identity_partitioned_p2_1 | i            | test_replica_identity_partitioned_p2_1_id_val_idx                              | t
+ test_replica_identity_partitioned_p2_2 | i            | test_replica_identity_schema.test_replica_identity_partitioned_p2_2_id_val_idx | t
+ test_replica_identity_partitioned_p2   | d            | test_replica_identity_partitioned_p2_id_val_idx                                | f
+ test_replica_identity_partitioned      | i            | test_replica_identity_partitioned_pkey                                         | t
+(5 rows)
+
+-- Compare the before and after to ensure replica identity preserved
+SELECT b.index_name, b.index_oid != c.oid AS rebuilt,
+       b.indisreplident != i.indisreplident AS ri_lost
+  FROM unnest(:'before_index_oids'::oid[],
+              :'before_index_names'::text[],
+              :'before_schema_names'::text[],
+              :'before_indisreplidents'::bool[])
+       AS b(index_oid, index_name, schema_name, indisreplident)
+  JOIN pg_namespace n ON n.nspname = b.schema_name
+  JOIN pg_class c ON c.relnamespace = n.oid AND c.relname = b.index_name
+  JOIN pg_index i ON c.oid = i.indexrelid
+  ORDER BY b.index_name;
+                    index_name                     | rebuilt | ri_lost 
+---------------------------------------------------+---------+---------
+ test_replica_identity_partitioned_p1_id_val_idx   | t       | f
+ test_replica_identity_partitioned_p2_1_id_val_idx | t       | f
+ test_replica_identity_partitioned_p2_2_id_val_idx | t       | f
+ test_replica_identity_partitioned_p2_id_val_idx   | t       | f
+ test_replica_identity_partitioned_pkey            | t       | f
+(5 rows)
+
+DROP SCHEMA test_replica_identity_schema CASCADE;
+NOTICE:  drop cascades to table test_replica_identity_schema.test_replica_identity_partitioned_p2_2
+DROP TABLE test_replica_identity_partitioned;
+-- A partition may use a partition index as replica identity even when its
+-- parent does not use the corresponding partitioned index.
+CREATE TABLE test_replica_identity_child_only (id int NOT NULL, val int NOT NULL) PARTITION BY RANGE (id);
+CREATE TABLE test_replica_identity_child_only_p1 PARTITION OF test_replica_identity_child_only
+    FOR VALUES FROM (0) TO (100);
+CREATE UNIQUE INDEX test_replica_identity_child_only_idx ON test_replica_identity_child_only (id, val);
+ALTER TABLE test_replica_identity_child_only_p1 REPLICA IDENTITY USING INDEX test_replica_identity_child_only_p1_id_val_idx;
+ALTER TABLE test_replica_identity_child_only ALTER COLUMN val TYPE bigint;
+SELECT tc.relname AS table_name, tc.relreplident, i.indexrelid::regclass AS index_name, i.indisreplident
+  FROM pg_class c
+  JOIN pg_index i ON c.oid = i.indexrelid
+  JOIN pg_class tc ON i.indrelid = tc.oid
+  WHERE tc.relname = 'test_replica_identity_child_only_p1' AND i.indisreplident;
+             table_name              | relreplident |                   index_name                   | indisreplident 
+-------------------------------------+--------------+------------------------------------------------+----------------
+ test_replica_identity_child_only_p1 | i            | test_replica_identity_child_only_p1_id_val_idx | t
+(1 row)
+
+DROP TABLE test_replica_identity_child_only;
+--
+-- Cleanup
+--
 DROP TABLE test_replica_identity;
 DROP TABLE test_replica_identity2;
 DROP TABLE test_replica_identity3;
diff --git a/src/test/regress/sql/replica_identity.sql b/src/test/regress/sql/replica_identity.sql
index 30daec05b71..4c0b1d246f7 100644
--- a/src/test/regress/sql/replica_identity.sql
+++ b/src/test/regress/sql/replica_identity.sql
@@ -134,6 +134,78 @@ ALTER TABLE test_replica_identity5 ALTER b SET NOT NULL;
 ALTER TABLE test_replica_identity5 DROP CONSTRAINT test_replica_identity5_pkey;
 ALTER TABLE test_replica_identity5 ALTER b DROP NOT NULL;
 
+--
+-- Test index rebuild preserves replica identity against partitioned tables
+--
+CREATE TABLE test_replica_identity_partitioned (id int NOT NULL, val int NOT NULL) PARTITION BY RANGE (id);
+CREATE TABLE test_replica_identity_partitioned_p1 PARTITION OF test_replica_identity_partitioned
+    FOR VALUES FROM (0) TO (100);
+CREATE TABLE test_replica_identity_partitioned_p2 PARTITION OF test_replica_identity_partitioned
+    FOR VALUES FROM (101) TO (200) PARTITION BY LIST (id);
+CREATE TABLE test_replica_identity_partitioned_p2_1 PARTITION OF test_replica_identity_partitioned_p2
+    FOR VALUES IN (150, 160);
+-- For better coverage, create a partition in a different schema
+CREATE SCHEMA test_replica_identity_schema;
+CREATE TABLE test_replica_identity_schema.test_replica_identity_partitioned_p2_2 PARTITION OF test_replica_identity_partitioned_p2
+    FOR VALUES IN (170, 180);
+CREATE UNIQUE INDEX test_replica_identity_partitioned_pkey ON test_replica_identity_partitioned (id, val);
+ALTER TABLE test_replica_identity_partitioned REPLICA IDENTITY USING INDEX test_replica_identity_partitioned_pkey;
+ALTER TABLE test_replica_identity_partitioned_p1 REPLICA IDENTITY USING INDEX test_replica_identity_partitioned_p1_id_val_idx;
+ALTER TABLE test_replica_identity_partitioned_p2_1 REPLICA IDENTITY USING INDEX test_replica_identity_partitioned_p2_1_id_val_idx;
+ALTER TABLE test_replica_identity_schema.test_replica_identity_partitioned_p2_2 REPLICA IDENTITY USING INDEX test_replica_identity_partitioned_p2_2_id_val_idx;
+SELECT tc.relname as table_name, tc.relreplident, i.indexrelid::regclass AS index_name, i.indisreplident FROM pg_class c
+  LEFT JOIN pg_index i ON c.oid = i.indexrelid
+  LEFT JOIN pg_class tc ON i.indrelid = tc.oid
+  WHERE c.relname LIKE 'test_replica_identity_partitioned%' and c.relkind in ('i', 'I')
+  ORDER BY c.relname;
+-- Store the index state before rebuild
+SELECT array_agg(c.oid ORDER BY c.relname) AS index_oids,
+       array_agg(c.relname::text ORDER BY c.relname) AS index_names,
+       array_agg(n.nspname ORDER BY c.relname) AS schema_names,
+       array_agg(i.indisreplident ORDER BY c.relname) AS indisreplidents
+  FROM pg_class c
+  JOIN pg_namespace n ON c.relnamespace = n.oid
+  JOIN pg_index i ON c.oid = i.indexrelid
+  WHERE c.relname LIKE 'test_replica_identity_partitioned%' AND c.relkind IN ('i', 'I') \gset before_
+ALTER TABLE test_replica_identity_partitioned ALTER COLUMN val TYPE bigint;
+SELECT tc.relname as table_name, tc.relreplident, i.indexrelid::regclass AS index_name, i.indisreplident FROM pg_class c
+  LEFT JOIN pg_index i ON c.oid = i.indexrelid
+  LEFT JOIN pg_class tc ON i.indrelid = tc.oid
+  WHERE c.relname LIKE 'test_replica_identity_partitioned%' and c.relkind in ('i', 'I')
+  ORDER BY c.relname;
+-- Compare the before and after to ensure replica identity preserved
+SELECT b.index_name, b.index_oid != c.oid AS rebuilt,
+       b.indisreplident != i.indisreplident AS ri_lost
+  FROM unnest(:'before_index_oids'::oid[],
+              :'before_index_names'::text[],
+              :'before_schema_names'::text[],
+              :'before_indisreplidents'::bool[])
+       AS b(index_oid, index_name, schema_name, indisreplident)
+  JOIN pg_namespace n ON n.nspname = b.schema_name
+  JOIN pg_class c ON c.relnamespace = n.oid AND c.relname = b.index_name
+  JOIN pg_index i ON c.oid = i.indexrelid
+  ORDER BY b.index_name;
+DROP SCHEMA test_replica_identity_schema CASCADE;
+DROP TABLE test_replica_identity_partitioned;
+
+-- A partition may use a partition index as replica identity even when its
+-- parent does not use the corresponding partitioned index.
+CREATE TABLE test_replica_identity_child_only (id int NOT NULL, val int NOT NULL) PARTITION BY RANGE (id);
+CREATE TABLE test_replica_identity_child_only_p1 PARTITION OF test_replica_identity_child_only
+    FOR VALUES FROM (0) TO (100);
+CREATE UNIQUE INDEX test_replica_identity_child_only_idx ON test_replica_identity_child_only (id, val);
+ALTER TABLE test_replica_identity_child_only_p1 REPLICA IDENTITY USING INDEX test_replica_identity_child_only_p1_id_val_idx;
+ALTER TABLE test_replica_identity_child_only ALTER COLUMN val TYPE bigint;
+SELECT tc.relname AS table_name, tc.relreplident, i.indexrelid::regclass AS index_name, i.indisreplident
+  FROM pg_class c
+  JOIN pg_index i ON c.oid = i.indexrelid
+  JOIN pg_class tc ON i.indrelid = tc.oid
+  WHERE tc.relname = 'test_replica_identity_child_only_p1' AND i.indisreplident;
+DROP TABLE test_replica_identity_child_only;
+
+--
+-- Cleanup
+--
 DROP TABLE test_replica_identity;
 DROP TABLE test_replica_identity2;
 DROP TABLE test_replica_identity3;
diff --git a/src/backend/commands/tablecmds.c b/src/backend/commands/tablecmds.c
index cb93c3e935a..4d667b31917 100644
--- a/src/backend/commands/tablecmds.c
+++ b/src/backend/commands/tablecmds.c
@@ -207,7 +207,8 @@ typedef struct AlteredTableInfo
 	List	   *changedConstraintDefs;	/* string definitions of same */
 	List	   *changedIndexOids;	/* OIDs of indexes to rebuild */
 	List	   *changedIndexDefs;	/* string definitions of same */
-	char	   *replicaIdentityIndex;	/* index to reset as REPLICA IDENTITY */
+	List	   *replicaIdentityIndexOids;	/* OIDs of index to reset as
+											 * REPLICA IDENTITY */
 	char	   *clusterOnIndex; /* index to use for CLUSTER */
 	List	   *changedStatisticsOids;	/* OIDs of statistics to rebuild */
 	List	   *changedStatisticsDefs;	/* string definitions of same */
@@ -692,10 +693,10 @@ static bool ATColumnChangeRequiresRewrite(Node *expr, AttrNumber varattno);
 static ObjectAddress ATExecAlterColumnType(AlteredTableInfo *tab, Relation rel,
 										   AlterTableCmd *cmd, LOCKMODE lockmode);
 static void RememberAllDependentForRebuilding(AlteredTableInfo *tab, AlterTableType subtype,
-											  Relation rel, AttrNumber attnum, const char *colName);
-static void RememberWholeRowDependentForRebuilding(AlteredTableInfo *tab, AlterTableType subtype, Relation rel);
-static void RememberConstraintForRebuilding(Oid conoid, AlteredTableInfo *tab);
-static void RememberIndexForRebuilding(Oid indoid, AlteredTableInfo *tab);
+											  Relation rel, AttrNumber attnum, const char *colName, LOCKMODE lockmode);
+static void RememberWholeRowDependentForRebuilding(AlteredTableInfo *tab, AlterTableType subtype, Relation rel, LOCKMODE lockmode);
+static void RememberConstraintForRebuilding(Oid conoid, AlteredTableInfo *tab, LOCKMODE lockmode);
+static void RememberIndexForRebuilding(Oid indoid, AlteredTableInfo *tab, LOCKMODE lockmode);
 static void RememberStatisticsForRebuilding(Oid stxoid, AlteredTableInfo *tab);
 static void ATPostAlterTypeCleanup(List **wqueue, AlteredTableInfo *tab,
 								   LOCKMODE lockmode);
@@ -8815,14 +8816,14 @@ ATExecSetExpression(AlteredTableInfo *tab, Relation rel, const char *colName,
 	 * Find everything that depends on the column (constraints, indexes, etc),
 	 * and record enough information to let us recreate the objects.
 	 */
-	RememberAllDependentForRebuilding(tab, AT_SetExpression, rel, attnum, colName);
+	RememberAllDependentForRebuilding(tab, AT_SetExpression, rel, attnum, colName, lockmode);
 
 	/*
 	 * Find whole-row referenced objects that depend on the column
 	 * (constraints, indexes, etc.), and record enough information to let us
 	 * recreate the objects.
 	 */
-	RememberWholeRowDependentForRebuilding(tab, AT_SetExpression, rel);
+	RememberWholeRowDependentForRebuilding(tab, AT_SetExpression, rel, lockmode);
 
 	/*
 	 * Drop the dependency records of the GENERATED expression, in particular
@@ -15375,7 +15376,7 @@ ATExecAlterColumnType(AlteredTableInfo *tab, Relation rel,
 	 * the info before executing ALTER TYPE, though, else the deparser will
 	 * get confused.
 	 */
-	RememberAllDependentForRebuilding(tab, AT_AlterColumnType, rel, attnum, colName);
+	RememberAllDependentForRebuilding(tab, AT_AlterColumnType, rel, attnum, colName, lockmode);
 
 	/*
 	 * Now scan for dependencies of this column on other things.  The only
@@ -15578,7 +15579,7 @@ ATExecAlterColumnType(AlteredTableInfo *tab, Relation rel,
  */
 static void
 RememberAllDependentForRebuilding(AlteredTableInfo *tab, AlterTableType subtype,
-								  Relation rel, AttrNumber attnum, const char *colName)
+								  Relation rel, AttrNumber attnum, const char *colName, LOCKMODE lockmode)
 {
 	Relation	depRel;
 	ScanKeyData key[3];
@@ -15624,7 +15625,7 @@ RememberAllDependentForRebuilding(AlteredTableInfo *tab, AlterTableType subtype,
 						relKind == RELKIND_PARTITIONED_INDEX)
 					{
 						Assert(foundObject.objectSubId == 0);
-						RememberIndexForRebuilding(foundObject.objectId, tab);
+						RememberIndexForRebuilding(foundObject.objectId, tab, lockmode);
 					}
 					else if (relKind == RELKIND_SEQUENCE)
 					{
@@ -15645,7 +15646,7 @@ RememberAllDependentForRebuilding(AlteredTableInfo *tab, AlterTableType subtype,
 
 			case ConstraintRelationId:
 				Assert(foundObject.objectSubId == 0);
-				RememberConstraintForRebuilding(foundObject.objectId, tab);
+				RememberConstraintForRebuilding(foundObject.objectId, tab, lockmode);
 				break;
 
 			case ProcedureRelationId:
@@ -15798,6 +15799,41 @@ RememberAllDependentForRebuilding(AlteredTableInfo *tab, AlterTableType subtype,
 	table_close(depRel, NoLock);
 }
 
+static void
+find_partition_replica_identity_indexes(AlteredTableInfo *tab, Oid relId, Oid indexId, LOCKMODE lockmode)
+{
+	List	   *partRelIds = NIL;
+
+	partRelIds = find_inheritance_children(relId, lockmode);
+	foreach_oid(partRelOid, partRelIds)
+	{
+		Relation	partRel;
+		Oid			partIndexId;
+
+		/* find_inheritance_children already got lock */
+		partRel = relation_open(partRelOid, NoLock);
+
+		partIndexId = index_get_partition(partRel, indexId);
+		if (!OidIsValid(partIndexId))
+		{
+			/* if we won't touch the partition, then unlock it */
+			relation_close(partRel, lockmode);
+			continue;
+		}
+
+		if (get_index_isreplident(partIndexId))
+			tab->replicaIdentityIndexOids = lappend_oid(tab->replicaIdentityIndexOids, partIndexId);
+
+		if (partRel->rd_rel->relkind == RELKIND_PARTITIONED_TABLE)
+		{
+			find_partition_replica_identity_indexes(tab, partRelOid, partIndexId, lockmode);
+		}
+
+		relation_close(partRel, NoLock);
+	}
+	list_free(partRelIds);
+}
+
 /*
  * Record information about dependencies between objects with whole-row Var
  * references (indexes, check constraints, etc.) and the relation.
@@ -15808,7 +15844,7 @@ RememberAllDependentForRebuilding(AlteredTableInfo *tab, AlterTableType subtype,
  * This function currently applies only to ALTER COLUMN SET EXPRESSION.
  */
 static void
-RememberWholeRowDependentForRebuilding(AlteredTableInfo *tab, AlterTableType subtype, Relation rel)
+RememberWholeRowDependentForRebuilding(AlteredTableInfo *tab, AlterTableType subtype, Relation rel, LOCKMODE lockmode)
 {
 	ScanKeyData skey;
 	Relation	pg_constraint;
@@ -15875,7 +15911,7 @@ RememberWholeRowDependentForRebuilding(AlteredTableInfo *tab, AlterTableType sub
 				 */
 				if (bms_is_member(InvalidAttrNumber - FirstLowInvalidHeapAttributeNumber, expr_attrs))
 				{
-					RememberConstraintForRebuilding(conform->oid, tab);
+					RememberConstraintForRebuilding(conform->oid, tab, lockmode);
 				}
 			}
 		}
@@ -15929,7 +15965,7 @@ RememberWholeRowDependentForRebuilding(AlteredTableInfo *tab, AlterTableType sub
 			 */
 			if (bms_is_member(InvalidAttrNumber - FirstLowInvalidHeapAttributeNumber, expr_attrs))
 			{
-				RememberIndexForRebuilding(index->indexrelid, tab);
+				RememberIndexForRebuilding(index->indexrelid, tab, lockmode);
 				continue;
 			}
 		}
@@ -15957,7 +15993,7 @@ RememberWholeRowDependentForRebuilding(AlteredTableInfo *tab, AlterTableType sub
 			 */
 			if (bms_is_member(InvalidAttrNumber - FirstLowInvalidHeapAttributeNumber, expr_attrs))
 			{
-				RememberIndexForRebuilding(index->indexrelid, tab);
+				RememberIndexForRebuilding(index->indexrelid, tab, lockmode);
 			}
 		}
 	}
@@ -15971,15 +16007,28 @@ RememberWholeRowDependentForRebuilding(AlteredTableInfo *tab, AlterTableType sub
  * needs to be reset.
  */
 static void
-RememberReplicaIdentityForRebuilding(Oid indoid, AlteredTableInfo *tab)
+RememberReplicaIdentityForRebuilding(Oid indoid, AlteredTableInfo *tab, LOCKMODE lockmode)
 {
-	if (!get_index_isreplident(indoid))
-		return;
+	if (get_index_isreplident(indoid))
+	{
+		foreach_oid(indexId, tab->replicaIdentityIndexOids)
+		{
+			if (IndexGetRelation(indexId, false) == tab->relid)
+				elog(ERROR, "relation %u has multiple indexes marked as replica identity", tab->relid);
+		}
 
-	if (tab->replicaIdentityIndex)
-		elog(ERROR, "relation %u has multiple indexes marked as replica identity", tab->relid);
+		tab->replicaIdentityIndexOids = lappend_oid(tab->replicaIdentityIndexOids, indoid);
+	}
 
-	tab->replicaIdentityIndex = get_rel_name(indoid);
+	/* For regular tables, we can only have one replica identity index */
+	if (tab->rel->rd_rel->relkind != RELKIND_PARTITIONED_TABLE)
+		return;
+
+	/*
+	 * A partition may use its partition index as replica identity even when
+	 * the parent does not, so always inspect the partition index hierarchy.
+	 */
+	find_partition_replica_identity_indexes(tab, tab->relid, indoid, lockmode);
 }
 
 /*
@@ -16002,7 +16051,7 @@ RememberClusterOnForRebuilding(Oid indoid, AlteredTableInfo *tab)
  * to be rebuilt (which we might already know).
  */
 static void
-RememberConstraintForRebuilding(Oid conoid, AlteredTableInfo *tab)
+RememberConstraintForRebuilding(Oid conoid, AlteredTableInfo *tab, LOCKMODE lockmode)
 {
 	/*
 	 * This de-duplication check is critical for two independent reasons: we
@@ -16047,7 +16096,7 @@ RememberConstraintForRebuilding(Oid conoid, AlteredTableInfo *tab)
 		indoid = get_constraint_index(conoid);
 		if (OidIsValid(indoid))
 		{
-			RememberReplicaIdentityForRebuilding(indoid, tab);
+			RememberReplicaIdentityForRebuilding(indoid, tab, lockmode);
 			RememberClusterOnForRebuilding(indoid, tab);
 		}
 	}
@@ -16058,7 +16107,7 @@ RememberConstraintForRebuilding(Oid conoid, AlteredTableInfo *tab)
  * to be rebuilt (which we might already know).
  */
 static void
-RememberIndexForRebuilding(Oid indoid, AlteredTableInfo *tab)
+RememberIndexForRebuilding(Oid indoid, AlteredTableInfo *tab, LOCKMODE lockmode)
 {
 	/*
 	 * This de-duplication check is critical for two independent reasons: we
@@ -16081,7 +16130,7 @@ RememberIndexForRebuilding(Oid indoid, AlteredTableInfo *tab)
 
 		if (OidIsValid(conoid))
 		{
-			RememberConstraintForRebuilding(conoid, tab);
+			RememberConstraintForRebuilding(conoid, tab, lockmode);
 		}
 		else
 		{
@@ -16098,7 +16147,7 @@ RememberIndexForRebuilding(Oid indoid, AlteredTableInfo *tab)
 			 * or if it is a clustered index, so that ATPostAlterTypeCleanup()
 			 * can queue up commands necessary to restore those properties.
 			 */
-			RememberReplicaIdentityForRebuilding(indoid, tab);
+			RememberReplicaIdentityForRebuilding(indoid, tab, lockmode);
 			RememberClusterOnForRebuilding(indoid, tab);
 		}
 	}
@@ -16277,19 +16326,29 @@ ATPostAlterTypeCleanup(List **wqueue, AlteredTableInfo *tab, LOCKMODE lockmode)
 	/*
 	 * Queue up command to restore replica identity index marking
 	 */
-	if (tab->replicaIdentityIndex)
+	if (tab->replicaIdentityIndexOids != NIL)
 	{
-		AlterTableCmd *cmd = makeNode(AlterTableCmd);
-		ReplicaIdentityStmt *subcmd = makeNode(ReplicaIdentityStmt);
-
-		subcmd->identity_type = REPLICA_IDENTITY_INDEX;
-		subcmd->name = tab->replicaIdentityIndex;
-		cmd->subtype = AT_ReplicaIdentity;
-		cmd->def = (Node *) subcmd;
-
-		/* do it after indexes and constraints */
-		tab->subcmds[AT_PASS_OLD_CONSTR] =
-			lappend(tab->subcmds[AT_PASS_OLD_CONSTR], cmd);
+		foreach_oid(indexId, tab->replicaIdentityIndexOids)
+		{
+			Oid			relId = IndexGetRelation(indexId, false);
+			Relation	partrel;
+			AlteredTableInfo *parttab;
+
+			AlterTableCmd *cmd = makeNode(AlterTableCmd);
+			ReplicaIdentityStmt *subcmd = makeNode(ReplicaIdentityStmt);
+
+			subcmd->identity_type = REPLICA_IDENTITY_INDEX;
+			subcmd->name = get_rel_name(indexId);
+			cmd->subtype = AT_ReplicaIdentity;
+			cmd->def = (Node *) subcmd;
+
+			/* do it after indexes and constraints */
+			partrel = relation_open(relId, lockmode);
+			parttab = ATGetQueueEntry(wqueue, partrel);
+			parttab->subcmds[AT_PASS_OLD_CONSTR] =
+				lappend(parttab->subcmds[AT_PASS_OLD_CONSTR], cmd);
+			relation_close(partrel, NoLock);
+		}
 	}
 
 	/*
-- 
2.50.1 (Apple Git-155)

